@@ -22,6 +22,9 @@
 #include "common_startup.hh"
 #include "ws-auth.hh"
 #include "secpoll-auth.hh"
+#include <sys/time.h>
+#include <sys/resource.h>
+#include <boost/foreach.hpp>
 
 bool g_anyToTcp;
 typedef Distributor<DNSPacket,DNSPacket,PacketHandler> DNSDistributor;
@@ -35,6 +38,7 @@ CommunicatorClass Communicator;
 UDPNameserver *N;
 int avg_latency;
 TCPNameserver *TN;
+vector<DNSDistributor*> g_distributors;
 
 ArgvMap &arg()
 {
@@ -94,7 +98,8 @@ void declareArguments()
   ::arg().set("allow-axfr-ips","Allow zonetransfers only to these subnets")="127.0.0.0/8,::1";
   ::arg().set("only-notify", "Only send AXFR NOTIFY to these IP addresses or netmasks")="0.0.0.0/0,::/0";
   ::arg().set("also-notify", "When notifying a domain, also notify these nameservers")="";
-  ::arg().set("slave-cycle-interval","Reschedule failed SOA serial checks once every .. seconds")="60";
+  ::arg().set("allow-notify-from","Allow AXFR NOTIFY from these IP ranges. If empty, drop all incoming notifies.")="0.0.0.0/0,::/0";
+  ::arg().set("slave-cycle-interval","Schedule slave freshness checks once every .. seconds")="60";
 
   ::arg().set("tcp-control-address","If set, PowerDNS can be controlled over TCP on this address")="";
   ::arg().set("tcp-control-port","If set, PowerDNS can be controlled over TCP on this address")="53000";
@@ -164,6 +169,55 @@ void declareArguments()
   ::arg().set("security-poll-suffix","Domain name from which to query security update notifications")="secpoll.powerdns.com.";
 }
 
+static uint64_t uptimeOfProcess(const std::string& str)
+{
+  static time_t start=time(0);
+  return time(0) - start;
+}
+
+static uint64_t getSysUserTimeMsec(const std::string& str)
+{
+  struct rusage ru;
+  getrusage(RUSAGE_SELF, &ru);
+
+  if(str=="sys-msec") {
+    return (ru.ru_stime.tv_sec*1000ULL + ru.ru_stime.tv_usec/1000);
+  }
+  else
+    return (ru.ru_utime.tv_sec*1000ULL + ru.ru_utime.tv_usec/1000);
+
+}
+
+static uint64_t getQCount(const std::string& str)
+try
+{
+  int totcount=0;
+  BOOST_FOREACH(DNSDistributor* d, g_distributors) {
+    if(!d)
+      continue;
+    int qcount, acount;
+    
+    d->getQueueSizes(qcount, acount);  // this does locking and other things, so don't get smart
+    totcount+=qcount;
+  }
+  return totcount;
+}
+catch(std::exception& e)
+{
+  L<<Logger::Error<<"Had error retrieving queue sizes: "<<e.what()<<endl;
+  return 0;
+}
+catch(PDNSException& e)
+{
+  L<<Logger::Error<<"Had error retrieving queue sizes: "<<e.reason<<endl;
+  return 0;
+}
+
+static uint64_t getLatency(const std::string& str) 
+{
+  return avg_latency;
+}
+
 void declareStats(void)
 {
   S.declare("udp-queries","Number of UDP queries received");
@@ -181,11 +235,11 @@ void declareStats(void)
   S.declare("recursing-answers","Number of recursive answers sent out");
   S.declare("recursing-questions","Number of questions sent to recursor");
   S.declare("corrupt-packets","Number of corrupt packets received");
-
+  S.declare("signatures", "Number of DNSSEC signatures made");
   S.declare("tcp-queries","Number of TCP queries received");
   S.declare("tcp-answers","Number of answers sent out over TCP");
 
-  S.declare("qsize-q","Number of questions waiting for database attention");
+  S.declare("qsize-q","Number of questions waiting for database attention", getQCount);
 
   S.declare("deferred-cache-inserts","Amount of cache inserts that were deferred because of maintenance");
   S.declare("deferred-cache-lookup","Amount of cache lookups that were deferred because of maintenance");
@@ -198,8 +252,15 @@ void declareStats(void)
   S.declare("dnsupdate-refused", "DNS update packets that are refused.");
   S.declare("dnsupdate-changes", "DNS update changes to records in total.");
 
+  S.declare("uptime", "Uptime of process in seconds", uptimeOfProcess);
+  S.declare("sys-msec", "Number of msec spent in system time", getSysUserTimeMsec);
+  S.declare("user-msec", "Number of msec spent in user time", getSysUserTimeMsec);
+  S.declare("meta-cache-size", "Number of entries in the metadata cache", DNSSECKeeper::dbdnssecCacheSizes);
+  S.declare("key-cache-size", "Number of entries in the key cache", DNSSECKeeper::dbdnssecCacheSizes);
+  S.declare("signature-cache-size", "Number of entries in the signature cache", signatureCacheSize);
+
   S.declare("servfail-packets","Number of times a server-failed packet was sent out");
-  S.declare("latency","Average number of microseconds needed to answer a question");
+  S.declare("latency","Average number of microseconds needed to answer a question", getLatency);
   S.declare("timedout-packets","Number of packets which weren't answered within timeout set");
   S.declare("security-status", "Security status based on regular polling");
   S.declareRing("queries","UDP Queries Received");
@@ -211,10 +272,8 @@ void declareStats(void)
   S.declareRing("remotes","Remote server IP addresses");
   S.declareRing("remotes-unauth","Remote hosts querying domains for which we are not auth");
   S.declareRing("remotes-corrupt","Remote hosts sending corrupt packets");
-
 }
 
-
 int isGuarded(char **argv)
 {
   char *p=strstr(argv[0],"-instance");
@@ -230,8 +289,7 @@ void sendout(const AnswerData<DNSPacket> &AD)
   N->send(AD.A);
 
   int diff=AD.A->d_dt.udiff();
-  avg_latency=(int)(1023*avg_latency/1024+diff/1024);
-
+  avg_latency=(int)(0.999*avg_latency+0.001*diff);
   delete AD.A;  
 }
 
@@ -240,21 +298,22 @@ void *qthread(void *number)
 {
   DNSPacket *P;
   DNSDistributor *distributor = DNSDistributor::Create(::arg().asNum("distributor-threads", 1)); // the big dispatcher!
+  int num = (int)(unsigned long)number;
+  g_distributors[num] = distributor;
   DNSPacket question;
   DNSPacket cached;
 
-  unsigned int &numreceived=*S.getPointer("udp-queries");
-  unsigned int &numreceiveddo=*S.getPointer("udp-do-queries");
+  AtomicCounter &numreceived=*S.getPointer("udp-queries");
+  AtomicCounter &numreceiveddo=*S.getPointer("udp-do-queries");
 
-  unsigned int &numreceived4=*S.getPointer("udp4-queries");
+  AtomicCounter &numreceived4=*S.getPointer("udp4-queries");
 
-  unsigned int &numreceived6=*S.getPointer("udp6-queries");
+  AtomicCounter &numreceived6=*S.getPointer("udp6-queries");
 
   int diff;
   bool logDNSQueries = ::arg().mustDo("log-dns-queries");
   bool doRecursion = ::arg().mustDo("recursor");
   bool skipfirst=true;
-  unsigned int maintcount = 0;
   UDPNameserver *NS = N;
 
   // If we have SO_REUSEPORT then create a new port for all receiver threads
@@ -270,23 +329,15 @@ void *qthread(void *number)
   }
 
   for(;;) {
+    if(!(P=NS->receive(&question))) { // receive a packet         inline
+      continue;                    // packet was broken, try again
+    }
+
     if (skipfirst)
       skipfirst=false;
     else  
       numreceived++;
 
-    if(number==0) { // only run on main thread
-      if(!((maintcount++)%250)) { // maintenance tasks - this can conceivably run infrequently
-        S.set("latency",(int)avg_latency);
-        int qcount, acount;
-        distributor->getQueueSizes(qcount, acount);  // this does locking and other things, so don't get smart
-        S.set("qsize-q",qcount);
-      }
-    }
-
-    if(!(P=NS->receive(&question))) { // receive a packet         inline
-      continue;                    // packet was broken, try again
-    }
 
     if(P->d_remote.getSocklen()==sizeof(sockaddr_in))
       numreceived4++;
@@ -350,6 +401,20 @@ void *qthread(void *number)
   return 0;
 }
 
+static void* dummyThread(void *)
+{
+  void* ignore=0;
+  pthread_exit(ignore);
+}
+
+static void triggerLoadOfLibraries()
+{
+  pthread_t tid;
+  pthread_create(&tid, 0, dummyThread, 0);
+  void* res;
+  pthread_join(tid, &res);
+}
+
 void mainthread()
 {
   Utility::srandom(time(0));
@@ -366,9 +431,13 @@ void mainthread()
    DNSPacket::s_udpTruncationThreshold = std::max(512, ::arg().asNum("udp-truncation-threshold"));
    DNSPacket::s_doEDNSSubnetProcessing = ::arg().mustDo("edns-subnet-processing");
 
-   doSecPoll(true); // this must be BEFORE chroot
+   try {
+     doSecPoll(true); // this must be BEFORE chroot
+   }
+   catch(...) {}
 
    if(!::arg()["chroot"].empty()) {  
+     triggerLoadOfLibraries();
      if(::arg().mustDo("master") || ::arg().mustDo("slave"))
         gethostbyname("a.root-servers.net"); // this forces all lookup libraries to be loaded
      Utility::dropGroupPrivs(newuid, newgid);
@@ -404,13 +473,14 @@ void mainthread()
   if(TN)
     TN->go(); // tcp nameserver launch
 
-  pthread_create(&qtid,0,carbonDumpThread, 0); // runs even w/o carbon, might change @ runtime    
-
   //  fork(); (this worked :-))
   unsigned int max_rthreads= ::arg().asNum("receiver-threads", 1);
+  g_distributors.resize(max_rthreads);
   for(unsigned int n=0; n < max_rthreads; ++n)
     pthread_create(&qtid,0,qthread, reinterpret_cast<void *>(n)); // receives packets
 
+  pthread_create(&qtid,0,carbonDumpThread, 0); // runs even w/o carbon, might change @ runtime    
+
   for(;;) {
     sleep(1800);
     try {
